{{
   Retargetable Concurrent Smallc Compiler.
   The RCSC generates P-codes for a virtual machine with two registers and a stack.
   This object interprets these P-codes and updates the registers and stack accordingly.
   Standard input and output are mapped to a serial driver.
   
}}


CON

  {{ comparison attributes }}
  EQ   =  $01   {{ 0x01 equal }}
  GE   =  $9C   {{ 0x9C greater or equal }}
  GEu  =  $1C   {{ 0x1C greater or equal (unsigned) }}
  GT   =  $8C   {{ 0x8C greater than }}
  GTu  =  $0C   {{ 0x0C greater than (unsigned) }}
  LE   =  $9A   {{ 0x9A less or equal }}
  LEu  =  $1A   {{ 0x1A less or equal (unsigned) }}
  LT   =  $8A   {{ 0x8A less than }}
  LTu  =  $0A   {{ 0x0A less than (unsigned) }}
  NE   =  $28   {{ 0x28 not equal }}

  EQf  =  $01   {{ 0x01 equal }}
  GEf  =  $9C   {{ 0x9C greater or equal }}
  GTf  =  $8C   {{ 0x8C greater than }}
  LEf  =  $9A   {{ 0x9A less or equal }}
  LTf  =  $8A   {{ 0x8A less than }}
  NEf  =  $28   {{ 0x28 not equal }}

  {{ pcode followed by 0 operands }}
  PCODE_OPER0 = $00
  {{ non-rcsc pcodes, to call native functions, arguments are on stack }}
  pNOP     =  $00   {{ no operation }}
  pCCARGC  =  $01   {{ get argument count }}
  pPRINTF  =  $02   {{ print to standard output }}

  {{ rcsc generated P-codes, arguments in operands }}
  pADD12   =  $1E  {{ add sr to pr }}
  pADD21   =  $1F  {{ add pr to sr }}
  pAND12   =  $20  {{ AND sr to pr }}
  pANEG1   =  $21  {{ arith negate pr }}
  pASL12   =  $22  {{ arith shift left sr by pr into pr }}
  pASR12   =  $23  {{ arith shift right sr by pr into pr }}
  pCALL1   =  $24  {{ call function thru pr }}
  pCLI     =  $25  {{ interrupt disable }}
  pCOM1    =  $26  {{ ones complement pr }}
  pCSCINIT =  $27  {{ concurrent Small C Initialization }}
  pDBL1    =  $28  {{ double pr }}
  pDBL2    =  $29  {{ double sr }}
  pDIV12   =  $2A  {{ div pr by sr }}
  pDIV12u  =  $2B  {{ div pr by sr unsigned }}
  pENTER   =  $2C  {{ set stack frame on function entry }}
  pLNEG1   =  $2D  {{ logical negate pr }}
  pMOD12   =  $2E  {{ modulo pr by sr }}
  pMOD12u  =  $2F  {{ modulo pr by sr unsigned }}
  pMOVE21  =  $30  {{ move pr to sr }}
  pMUL12   =  $31  {{ multiply pr by sr }}
  pMUL12u  =  $32  {{ multiply pr by sr unsigned }}
  pOR12    =  $33  {{ OR sr onto pr }}
  pPOP2    =  $34  {{ pop stack into sr }}
  pPUSH1   =  $35  {{ push pr onto stack }}
  pPUSH2   =  $36  {{ push sr }}
  pPUTbp1  =  $37  {{ put pr byte in mem thru sr ptr }}
  pPUTwp1  =  $38  {{ put pr word in mem thru sr ptr }}
  pRETURN  =  $39  {{ restore stack and return }}
  pSTI     =  $3A  {{ interrupt enable }}
  pSUB12   =  $3B  {{ sub sr from pr }}
  pSWAP12  =  $3C  {{ swap pr and sr }}
  pSWAP1s  =  $3D  {{ swap pr and top of stack }}
  pSWITCH  =  $3E  {{ find switch case }}
  pXOR12   =  $3F  {{ XOR pr with sr }}

  {{ pcode followed by 1 operand }}
  PCODE_OPER1 = $40
  pADDbpn  =  $40  {{ add n to mem byte thru sr ptr }}
  pARGCNTn =  $41  {{ pass arg count to function }}
  pCMP12   =  $42  {{ set pr to (sr CMP pr) }}

  {{ pcode followed by 2 operands }}
  PCODE_OPER2 = $60
  pADD1n   =  $60  {{ add n to pr }}
  pADD2n   =  $61  {{ add immediate to sr }}
  pADDSP   =  $62  {{ add to stack pointer }}
  pADDwpn  =  $63  {{ add n to mem word thru sr ptr }}
  pANDbm1  =  $64  {{ bitwise and pr with mem byte thru label }}
  pANDwm1  =  $65  {{ bitwise and pr with mem word thru label }}
  pCALLm   =  $66  {{ call function directly }}
  pGETb1m  =  $67  {{ get byte into pr from mem thru label }}
  pGETb1mu =  $68  {{ get unsigned byte into pr from mem thru label }}
  pGETb1p  =  $69  {{ get byte into pr from mem thru sr ptr }}
  pGETb1pu =  $6A  {{ get unsigned byte into pr from mem thru sr ptr }}
  pGETb1s  =  $6B  {{ get byte into pr from stack }}
  pGETb1su =  $6C  {{ get unsigned byte into pr from stack }}
  pGETw1m  =  $6D  {{ get word into pr from mem thru label }}
  pGETw1n  =  $6E  {{ get word of value n into pr }}
  pGETw1p  =  $6F  {{ get word into pr from mem thru sr ptr }}
  pGETw1s  =  $70  {{ get word into pr from stack }}
  pGETw2m  =  $71  {{ get word into sr from mem (label) }}
  pGETw2n  =  $72  {{ get word of value n into sr }}
  pGETw2p  =  $73  {{ get word into sr thru sr ptr }}
  pGETw2s  =  $74  {{ get word into sr from stack }}
  pJMPm    =  $75  {{ jump to label }}
  pORbm1   =  $76  {{ bitwise or pr with mem byte thru label }}
  pORwm1   =  $77  {{ bitwise or pr with mem word thru label }}
  pPOINT1m =  $78  {{ point pr to mem item thru label }}
  pPOINT1s =  $79  {{ point pr to stack item }}
  pPOINT2m =  $7A  {{ point sr to mem thru label }}
  pPOINT2s =  $7B  {{ point sr to stack }}
  pPUSHm   =  $7C  {{ push word from mem thru label }}
  pPUSHp   =  $7D  {{ push word from mem thru sr ptr }}
  pPUSHs   =  $7E  {{ push word from stack }}
  pPUTbm1  =  $7F  {{ put pr byte in mem thru label }}
  pPUTwm1  =  $80  {{ put pr word in mem thru label }}
  pXORbm1  =  $81  {{ bitwise xor pr with mem byte thru label }}
  pXORwm1  =  $82  {{ bitwise xor pr with mem word thru label }}

  {{ pcode followed by 3 operands }}
  PCODE_OPER3 = $A0
  pADDbmn  =  $A0  {{ add n to mem byte thru label }}
  pANDbmn  =  $A1  {{ bitwise and n with mem byte thru label }}
  pCMP10f  =  $A2  {{ jump to label if (pr CMP 0) is false }}
  pORbmn   =  $A3  {{ bitwise or n with mem byte thru label }}
  pPUTbmn  =  $A4  {{ put n into mem byte thru label }}
  pXORbmn  =  $A5  {{ bitwise xor n with mem byte thru label }}

  {{ pcode followed by 4 operands }}
  PCODE_OPER4 = $C0
  pADDwmn  =  $C0  {{ add n to mem word thru label }}
  pANDwmn  =  $C1  {{ bitwise and n with mem word thru label }}
  pORwmn   =  $C2  {{ bitwise or n with mem word thru label }}
  pPUTwmn  =  $C3  {{ put n into mem word thru label }}
  pXORwmn  =  $C4  {{ bitwise xor n with mem word thru label }}


OBJ

  debug: "MultiCogSerialDebug"

  
DAT

  {{ math registers }}
  vmAXlong      long    0               {{ for multiply, divide and modulo }}
  vmBXlong      long    0               {{ for multiply, divide and modulo }}

  {{ the virtual machine registers }}
  vmAX          word    0               {{ primary register (pr)           }}
  vmBX          word    0               {{ secondary register (sr)         }}
  vmSP          word    0               {{ stack pointer register          }}

  {{ stack parameter registers }}
  vmStackBase   word    0               {{ base address of atack           }}
  vmStackSize   word    0               {{ size of stack in bytes          }}
  vmStackBottom word    0               {{ lowest value of vmSP            }}
  vmStackTop    word    0               {{ highest value of vmSP           }}
  vmMain        word    0               {{ main address of rcsc C code     }}
  
  {{ working registers }}
  vmCX          word    0               {{ auxiliary register #1           }}
  vmDX          word    0               {{ auxiliary register #2           }}
  vmEX          word    0               {{ auxiliary register #3           }}
  vmPC          word    0               {{ program counter                 }}
  vmBP          word    0               {{ stack frame register            }}
  vmPCODE       byte    0               {{ pseude code operator            }}
  vmARGCNT      byte    0               {{ argument count                  }}
  vmINTERRUPT   byte    0               {{ tue if interrupt enabled        }}


PUB reset(main,stack,size)
  '/**
  ' * Reset program counter and initialize stack.
  ' *
  ' * @param main Start address (main of C program).
  ' * @param stack Base of stack area.
  ' * @param size Size of stack area in bytes. 
  ' */
  vmMain := main          {{ save parameters }}
  vmStackBase := stack    
  vmStackSize := size
  vmSP := stack + size    {{initialize stack }}
  vmStackTop := vmSP
  vmStackBottom := stack
  vmPC := main            {{initialize program counter }}
  

PUB run(pcodes)
  '/**
  ' * Run a number of P-codes
  ' *
  ' * @param pcodes Number of P-codes to execute. Use 0 to run forever.
  ' */
  if pcodes > 0
    repeat pcodes
      nextPcode
  else
    repeat
      nextPcode
      

PUB getPR: pr
  '/**
  ' * Get primary register.
  ' *
  ' * @return Primary register.
  ' */
  return vmAX


PUB getSR: sr
  '/**
  ' * Get secondary register.
  ' *
  ' * @return Secondary register.
  ' */
  return vmBX


PUB getPC: pc
  '/**
  ' * Get program counter.
  ' *
  ' * @return Program counter.
  ' */
  return vmPC


PUB getSP: sp
  '/**
  ' * Get stack pointer.
  ' *
  ' * @return Stack pointer.
  ' */
  return vmSP


PUB stackOK: YesNo
  '/**
  ' * Check stack and memorize lowest and highest stack pointer values.
  ' *
  ' * @return True if stack pointer in valid area, false if stack pointer ever moved outside stack area.
  ' */
  if vmSP < vmSTACKbottom
    vmSTACKbottom := vmSP
  if vmSP > vmSTACKtop
    vmSTACKtop := vmSP
  return (vmStackBottom => vmStackBase) AND (vmStackTop =< vmStackBase + vmStackSize)


PUB getLowestSP: bottom
  '/**
  ' * Get lowest stack pointer value.
  ' *
  ' * @return Lowest stack pointer value.
  ' */
  return vmSTACKbottom
     
  
PUB getHighestSP: top
  '/**
  ' * Get highest stack pointer value.
  ' *
  ' * @return Highest stack pointer value.
  ' */
  return vmSTACKtop
     
  
PUB nextPcode | sign
  '/**
  ' * Execute next P-code.
  ' */
  vmPCODE := byte[vmPC++]              {{ pcode }}
  if vmPCODE => PCODE_OPER1
    vmCX.byte[0] := byte[vmPC++]       {{ operand 1 }}
    if vmPCODE => PCODE_OPER2
      vmCX.byte[1] := byte[vmPC++]     {{ operand 2 }}
      if vmPCODE => PCODE_OPER3
        vmDX.byte[0] := byte[vmPC++]   {{ operand 3 }}
        if vmPCODE => PCODE_OPER4
          vmDX.byte[1] := byte[vmPC++] {{ operand 4 }}
          {{ pcode operators with 4 operands }}
          case vmPCODE
            pADDwmn:  {{ add n to memory word thru label }}
                      {{ addr in CX, n in DX }}
                      vmEX.byte[0] := byte[vmCX]
                      vmEX.byte[1] := byte[vmCX+1]
                      vmDX += vmEX
                      byte[vmCX] := vmDX.byte[0]
                      byte[vmCX+1] := vmDX.byte[1]

            pANDwmn:  {{ bitwise and n with mem word thru label }}
                      {{ addr in CX, n in DX }}
                      byte[vmCX] &= vmDX.byte[0]
                      byte[vmCX+1] &= vmDX.byte[1]

            pORwmn:   {{ bitwise or n with mem word thru label }}
                      {{ addr in CX, n in DX }}
                      byte[vmCX] |= vmDX.byte[0]
                      byte[vmCX+1] |= vmDX.byte[1]

            pPUTwmn:  {{ put n into memory word thru label }}
                      {{ addr in CX, n in DX }}
                      byte[vmCX] := vmDX.byte[0]
                      byte[vmCX+1] := vmDX.byte[1]

            pXORwmn:  {{ bitwise xor n with mem word thru label }}
                      {{ addr in CX, n in DX }}
                      byte[vmCX] ^= vmDX.byte[0]
                      byte[vmCX+1] ^= vmDX.byte[1]

        else
          {{ pcode operators with 3 operands }}
          case vmPCODE
            pADDbmn:  {{ add n to memory byte thru label }}
                      {{ addr in CX, n in DL }}
                      byte[vmCX] += vmDX.byte[0]

            pANDbmn:  {{ bitwise and n with memory byte }}
                      {{ addr in CX, n in DL }}
                      byte[vmCX] &= vmDX.byte[0]

            pCMP10f:  {{ jump to label if (pr CMP 0) is false }}
                      {{ addr in CX, cmp in DL }}
                      case vmDX.byte[0]
                        EQf: {{ equal }}
                             vmEX := (vmAX == 0)
                        GEf: {{ greater or equal }}
                             vmEX := (vmAX => 0)
                        GTf: {{ greater }}
                             vmEX := (vmAX > 0)
                        LEf: {{ less or equal }}
                             vmEx := (vmAX =< 0)
                        LTf: {{ less }}
                             vmEX := (vmAX < 0)
                        NEf: {{ not equal }}
                             vmEX := (vmAX <> 0)
                        other: vmEX := true
                      if NOT vmEX
                        vmPC := vmCX

            pORbmn:   {{ bitwise or n with memory byte }}
                      {{ addr in CX, n in DL }}
                      byte[vmCX] |= vmDX.byte[0]

            pPUTbmn:  {{ put n into memory byte thru label }}
                      {{ addr in CX, n in DL }}
                      byte[vmCX] := vmDX.byte[0]

            pXORbmn:  {{ bitwise xor n with memory byte }}
                      {{ addr in CX, n in DL }}
                      byte[vmCX] ^= vmDX.byte[0]

      else
        {{ pcode operators with 2 operands }}
        case vmPCODE
          pADD1n:   {{ add n to pr }}
                    {{ n in CX }}
                    vmAX += vmCX

          pADD2n:   {{ add n to sr }}
                    {{ n in CX }}
                    vmBX += vmCX

          pADDSP:   {{ add n to stackpointer }}
                    {{ n in CX }}
                    vmSP += vmCX

          pADDwpn:  {{ add n to memory word thru sr pointer }}
                    {{ n in CX }}
                    vmEX.byte[0] := byte[vmBX]
                    vmEX.byte[1] := byte[vmBX+1]
                    vmCX += vmEX
                    byte[vmBX] := vmCX.byte[0]
                    byte[vmBX+1] := vmCX.byte[1]

          pANDbm1:  {{ bitwise and pr with mem byte thru label }}
                    {{ addr in CX }}
                    byte[vmCX] &= vmAX.byte[0]

          pANDwm1:  {{ bitwise and pr with mem word thru label }}
                    {{ addr in CX }}
                    byte[vmCX] &= vmAX.byte[0]
                    byte[vmCX+1] &= vmAX.byte[1]

          pCALLm:   {{ call function directly }}
                    {{ addr in CX }}
                    vmCX += vmMain
                    byte[--vmSP] := vmPC.byte[1]
                    byte[--vmSP] := vmPC.byte[0]
                    vmPC := vmCX

          pGETb1m:  {{ get byte into pr from memory thru label }}
                    {{ addr in CX }}
                    vmAX.byte[0] := byte[vmCX]
                    vmAX.byte[1] := (vmAX.byte[0] & $80) <> 0

          pGETb1mu: {{ get unsigned byte into pr from memory thru label }}
                    {{ addr in CX }}
                    vmAX.byte[0] := byte[vmCX]
                    vmAX.byte[1] := 0

          pGETb1p:  {{ get byte into pr from memory thru sr pointer }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBX+vmCX]
                    vmAX.byte[1] := (vmAX.byte[0] & $80) <> 0
                  
          pGETb1pu: {{ get unsigned byte into pr from memory thru sr pointer }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBX+vmCX]
                    vmAX.byte[1] := 0

          pGETb1s:  {{ get byte into pr from stack }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBP+vmCX]
                    vmAX.byte[1] := (vmAX.byte[0] & $80) <> 0
                  
          pGETb1su: {{ get unsigned byte into pr from stack }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBP+vmCX]
                    vmAX.byte[1] := 0

          pGETw1m:  {{ get word into pr from memory thru label }}
                    {{ addr in CX }}
                    vmAX.byte[0] := byte[vmCX]
                    vmAX.byte[1] := byte[vmCX+1]

          pGETw1n:  {{ get word of value n into pr }}
                    {{ n in CX }}
                    vmAX := vmCX

          pGETw1p:  {{ get word into pr from memory thru sr pointer }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBX+vmCX]
                    vmAX.byte[1] := byte[vmBX+vmCX+1]
                  
          pGETw1s:  {{ get word into pr from stack }}
                    {{ offset in CX }}
                    vmAX.byte[0] := byte[vmBP+vmCX]
                    vmAX.byte[1] := byte[vmBP+vmCX+1]
                  
          pGETw2m:  {{ get word into sr from memory thru label }}
                    {{ addr in CX }}
                    vmBX.byte[0] := byte[vmCX]
                    vmBX.byte[1] := byte[vmCX+1]

          pGETw2n:  {{ get word of value n into sr }}
                    {{ n in CX }}
                    vmBX := vmCX

          pGETw2p:  {{ get word into sr from memory thru sr pointer }}
                    {{ offset in CX }}
                    vmDX.byte[0] := byte[vmBX+vmCX]
                    vmDX.byte[1] := byte[vmBX+vmCX+1]
                    vmBX := vmDX
                  
          pGETw2s:  {{ get word into sr from stack }}
                    {{ offset in CX }}
                    vmBX.byte[0] := byte[vmBP+vmCX]
                    vmBX.byte[1] := byte[vmBP+vmCX+1]
                  
          pJMPm:    {{ jump to label }}
                    {{ addr in CX }}
                    vmCX += vmMain
                    vmPC := vmCX

          pORbm1:   {{ bitwise or pr with mem byte thru label }}
                    {{ addr in CX }}
                    byte[vmCX] |= vmAX.byte[0]
                    
          pORwm1:   {{ bitwise or pr with mem word thru label }}
                    {{ addr in CX }}
                    byte[vmCX] |= vmAX.byte[0]
                    byte[vmCX+1] |= vmAX.byte[1]

          pPOINT1m: {{ point pr to memory item thru label }}
                    {{ addr in CX }}
                    vmCX += vmMain
                    vmAX := vmCX
                    
          pPOINT1s: {{ point pr to stack item }}
                    {{ offset in CX }}
                    vmAX := vmBP+vmCX
                    
          pPOINT2m: {{ point sr to memory item thru label }}
                    {{ addr in CX }}
                    vmBX := vmCX
                    
          pPOINT2s: {{ point sr to stack item }}
                    {{ offset in CX }}
                    vmBX := vmBP+vmCX
                    
          pPUSHm:   {{ push word from memory thru label }}
                    {{ addr in CX }}
                    byte[--vmSP] := byte[vmCX]
                    byte[--vmSP] := byte[vmCX+1]
                    
          pPUSHp:   {{ push word from memory thru sr pointer }}
                    {{ offset in CX }}
                    byte[--vmSP] := byte[vmBX+vmCX]
                    byte[--vmSP] := byte[vmBX+vmCX+1]
                    
          pPUSHs:   {{ push word from stack }}
                    {{ offset in CX }}
                    byte[--vmSP] := byte[vmBP+vmCX]
                    byte[--vmSP] := byte[vmBP+vmCX+1]
                    
          pPUTbm1:  {{ put pr byte into memory thru label }}
                    {{ addr in CX }}
                    byte[vmCX] := vmAX.byte[0]
                    
          pPUTwm1:  {{ put pr word into memory thru label }}
                    {{ addr in CX }}
                    byte[vmCX] := vmAX.byte[0]
                    byte[vmCX+1] := vmAX.byte[1]
                    
          pXORbm1:  {{ bitwise xor pr with mem byte thru label }}
                    {{ addr in CX }}
                    byte[vmCX] ^= vmAX.byte[0]

          pXORwm1:  {{ bitwise xor pr with mem word thru label }}
                    {{ addr in CX }}
                    byte[vmCX] ^= vmAX.byte[0]
                    byte[vmCX+1] ^= vmAX.byte[1]

    else
      {{ pcode operators with 1 operand }}
      case vmPCODE
        pADDbpn:  {{ add n to memory byte through sr pointer }}
                  {{ n in CL }}
                  byte[vmBX] += vmCX.byte[0]
                  
        pARGCNTn: {{ pass argument count to function }}
                  {{ argcnt in CL }}
                  vmARGCNT := vmCX.byte[0]
                  
        pCMP12:   {{ set pr to (sr CMP pr) }}
                  {{ cmp in CL }}
                  case vmCX.byte[0]
                    EQ : {{ 0x01 equal }}
                         vmAX := (vmBX == vmAX)
                    GE : {{ 0x9C greater or equal }}
                         if (vmBX == vmAX)
                           vmAX := true
                         else
                           vmAX := ((vmBX - vmAX) & $8000) == 0 
                    GEu: {{ 0x1C greater or equal (unsigned) }}
                         vmAX := (vmBX => vmAX)
                    GT : {{ 0x8C greater than }}
                         vmAX := ((vmBX - vmAX) & $8000) == 0
                    GTu: {{ 0x0C greater than (unsigned) }}
                         vmAX := (vmAX > vmBX)
                    LE : {{ 0x9A less or equal }}
                         if (vmBX == vmAX)
                           vmAX := true
                         else
                           vmAX := ((vmBX - vmAX) & $8000) <> 0
                    LEu: {{ 0x1A less or equal (unsigned) }}
                         vmAX := (vmBX =< vmAX)
                    LT : {{ 0x8A less than }}
                         vmAX := ((vmBX - vmAX) & $8000) <> 0
                    LTu: {{ 0x0A less than (unsigned) }}
                         vmAX := (vmBX < vmAX)
                    NE : {{ 0x28 not equal }}
                         vmAX := (vmAX <> vmBX)
                       
  else
    {{ pcode operators with 0 operands }}
    case vmPCODE
      pNOP:     {{ no operation }}
                {{ do nothing }}
                
      pADD12:   {{ add sr to pr }}
                vmAX += vmBX
                
      pADD21:   {{ add pr to sr }}
                vmBX += vmAX
                
      pAND12:   {{ and pr with sr }}
                vmAX &= vmBX
                
      pANEG1:   {{ arithmetically negate pr }}
                vmAX := -vmAX
                
      pASL12:   {{ arithmetically shift left sr by pr into pr }}
                vmAX := (vmBX << vmAX)
                
      pASR12:   {{ arithmetically shift right sr by pr into pr }}
                vmAX := (vmBX -> vmAX)
                
      pCALL1:   {{ call function thru pr }}
                byte[--vmSP] := vmPC.byte[1]
                byte[--vmSP] := vmPC.byte[0]
                vmPC := vmAX
                
      pCCARGC:  {{ get argument count }}
                {{ argument count in vmARGCNT }}
                vmAX := vmARGCNT 

      pCLI:     {{ disable interrupt }}
                vmINTERRUPT := false
      
      pCOM1:    {{ one complement pr }}
                vmAX ^= $FFFF
                
      pCSCINIT: {{ initialize csc }}
      
      pDBL1:    {{ double pr }}
                vmAX <<= 1
                
      pDBL2:    {{ double sr }}
                vmBX <<= 1
                
      pDIV12:   {{ divide pr by sr }}
                vmAXlong := ||(~~vmAX) {{sign extend b15, then absolute value }}
                vmBXlong := ||(~~vmBX) {{sign extend b15, then absolute value }}
                sign := ((vmAX ^ vmBX) & $8000) == 0
                vmAXlong /= vmBXlong
                if sign
                  vmAXlong := -vmAXlong
                vmAX := vmAXlong.byte[0]

      pDIV12u:  {{ divide pr by sr (unsigned) }}
                vmAXlong := vmAX
                vmBXlong := vmBX
                vmAXlong /= vmBXlong
                vmAX := vmAXlong.byte[0]

      pENTER:   {{ set stack frame upon function entry }}
                byte[--vmSP] := vmBP.byte[1]
                byte[--vmSP] := vmBP.byte[0]
                vmBP := vmSP
                
      pLNEG1:   {{ logical negation of pr }}
                vmAX := NOT vmAX
                
      pMOD12:   {{ modulo pr by sr }}
                vmAXlong := ||(~~vmAX) {{sign extend b15, then absolute value }}
                vmBXlong := ||(~~vmBX) {{sign extend b15, then absolute value }}
                sign := ((vmAX ^ vmBX) & $8000) == 0
                vmAXlong //= vmBXlong
                if sign
                  vmAXlong := -vmAXlong
                vmAX := vmAXlong.byte[0]
                
      pMOD12u:  {{ modulo pr by sr (unsigned) }}
                vmAXlong := vmAX
                vmBXlong := vmBX
                vmAXlong //= vmBXlong
                vmAX := vmAXlong.byte[0]
      
      pMOVE21:  {{ move pr to sr }}
                vmBX := vmAX
                
      pMUL12:   {{ multiply pr by sr }}
                vmAXlong := ||(~~vmAX) {{sign extend b15, then absolute value }}
                vmBXlong := ||(~~vmBX) {{sign extend b15, then absolute value }}
                sign := ((vmAX ^ vmBX) & $8000) == 0
                vmAXlong *= vmBXlong
                if sign
                  vmAXlong := -vmAXlong
                vmAX := vmAXlong.byte[0]

      pMUL12u:  {{ multiply pr by sr (unsigned) }}
                vmAXlong := vmAX
                vmBXlong := vmBX
                vmAXlong *= vmBXlong
                vmAX := vmAXlong.byte[0]
      
      pNOP:     {{ no operation }}
                {{ do nothing }}

      pOR12:    {{ or pr with sr }}
                vmAX |= vmBX
                
      pPOP2:    {{ pop stack into sr }}
                vmBX.byte[0] := byte[vmSP++]
                vmBX.byte[1] := byte[vmSP++]
                
      pPRINTF:  {{ print to standard output }}
                {{ formatstring addr on stack BP+6/7, value on stack BP+4/5 }}
                vmCX.byte[1] := byte[vmBP+7] 
                vmCX.byte[0] := byte[vmBP+6]
                vmDX.byte[1] := byte[vmBP+5] 
                vmDX.byte[0] := byte[vmBP+4]
                debug.cprintf(vmCX,vmDX,false) 

      pPUSH1:   {{ push pr onto stack }}
                byte[--vmSP] := vmAX.byte[1]
                byte[--vmSP] := vmAX.byte[0]
                
      pPUSH2:   {{ push sr onto stack }}
                byte[--vmSP] := vmBX.byte[1]
                byte[--vmSP] := vmBX.byte[0]
                
      pPUTbp1:  {{ put pr byte into memory thru sr pointer }}
                byte[vmBX] := vmAX.byte[0]
                
      pPUTwp1:  {{ put pr word into memory thru sr pointer }}
                byte[vmBX] := vmAX.byte[0]
                byte[vmBX+1] := vmAX.byte[1]
                
      pRETURN:  {{ restore stack and return }}
                vmSP := vmBP
                vmBP.byte[0] := byte[vmSP++]
                vmBP.byte[1] := byte[vmSP++]
                vmPC.byte[0] := byte[vmSP++]
                vmPC.byte[1] := byte[vmSP++]
                
      pSTI:     {{ enable interrupt }}
                vmINTERRUPT := true
      
      pSUB12:   {{ subtract sr from pr }}
                vmAX -= vmBX
                
      pSWAP12:  {{ swap pr and sr }}
                vmCX := vmAX
                vmAX := vmBX
                vmBX := vmCX
                
      pSWAP1s:  {{ swap pr and top of stack }}
                vmCX.byte[0] := byte[vmSP++]
                vmCX.byte[1] := byte[vmSP++]
                byte[--vmSP] := vmAX.byte[1]
                byte[--vmSP] := vmAX.byte[0]
                vmAX := vmCX
                
      pSWITCH:  {{ scan switch table to find case }}
                {{ remember: AX=compare value, PC=table address upon entry }}
                {{ the table consists of address/value pairs, a zero address }}
                {{ terminates the table and is followed by a jump to the default prefix }}
                repeat
                  vmCX.byte[0] := byte[vmPC++]   {{ load jump address for case }}
                  vmCX.byte[1] := byte[vmPC++]
                  if vmCX <> 0                   {{ if addr == 0 then do default }}
                    vmDX.byte[0] := byte[vmPC++] {{ get case value }}
                    vmDX.byte[1] := byte[vmPC++]
                    if vmAX == vmDX              {{ compare with compare value }}
                      vmPC := vmCX
                      vmCX := 0
                until vmCX == 0
      
      pXOR12:   {{ xor pr with sr }}
                vmAX ^= vmBX

